#
#  $Id$
#
#  (C) Copyright 2003 Kai Sterker <kaisterker@linuxgames.com>
#  Part of the Adonthell Project http://adonthell.linuxgames.com
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY.
#
#  See the COPYING file for more details
#

#
# Class that implements a certain damage type and method. There are 
# four damage types corresponding to the four types of resistances
# (Physical, Fire, Poison and Magic), plus a fifth type that can be
# used to modify either of the former (especially useful for fighting
# Feats).
# 
# Damage objects that deal the same kind of damage need to have the
# same id, so that their effect can be accumulated. Damage objects
# that should not be accumulated should not get the same id.
#
# Note that this class is abstract and needs to be subclassed to
# be of any use.
#
class damage (object):
    # -- The different types of damage. These values are also used
    #    to define the priority of damage types. I.e., if some object
    #    causes both Physical and Fire damage, the Fire damage will
    #    always be applied first.
    PHYSICAL = 0
    FIRE = 1
    POISON = 2
    MAGIC = 3
    SPECIAL = 4
    
    # -- Create a new damage object. 
    #    d_id: the Id of the object, used to accumulate similar damage objects
    #    d_type: the type of damage (Physical, Fire, Poison, Magic, Special)
    #    d_dmg: the damage the object does, used by the deal_damage method
    def __init__ (self, d_id, d_type, d_dmg):
        self.Id = d_id
        self.Type = d_type
        self.Damage = d_dmg
    
    # -- Compare two damage object for their priority.
    def __cmp__ (self, other):
        return other.Type - self.Type
    
    # -- accumulate the damage of two similar damage objects. Override if 
    #    needed.
    def merge (self, other):
        if self.Id == other.Id and self.Type == other.Type:
            self.Damage = (self.Damage[0] + other.Damage[0], self.Damage[1] + other.Damage[1])
            return 1
        else: return 0
        
    # -- return a textual representation of the damage object. Must be
    #    implemented by any subclass.
    def to_string (self):
        raise NotImplementedError, "damage::to_string"
    
    # -- apply the damage effect to a character. Must be implemented by
    #    any subclass.
    def deal_damage (self, defender):
        raise NotImplementedError, "damage::deal_damage"

#
# A damage_chain is used to accumulate the damage a character, item, 
# spell rune, feat, etc. does.
#
class damage_chain (object):
    
    # -- create a new damage_chain
    #    func: damage function to use
    #    next: next object in damage chain
    def __init__ (self, func = None, next = None):
        self.__Func = func
        self.__Next = next
    
    # -- retrieve the damage dealt by this damage chain
    #    Equal types of damages are accumulated, different ones are
    #    returned as seperate damage objects. The resulting list is
    #    sorted, so that SPECIAL damage objects come first, then 
    #    MAGIC, POISON, FIRE and last PHYSICAL damage objects.
    def execute (self, d_list):
        if self.__Func != None:
            def merge (d_list, d_obj):
               for obj in d_list:
                  if obj.merge (d_obj): return
               d_list.append (d_obj)
            merge (d_list, self.__Func ())
        if self.__Next != None:
            return self.__Next.execute (d_list)
        else:
            d_list.sort ()
            return d_list

    # -- return a textual description of the damage chain
    def to_string (self):
        d_list = self.execute ([])
        d_list.reverse ()
        string = ""
        for d_obj in d_list:
            string = "%s%s\n" % (string, d_obj.to_string ())
        return string
        
    # -- assign a new damage function
    #    func: the damage function        
    def set_func (self, func):
        self.__Func = func
        
    # -- insert a new damage object after this object
    #    next: the damage object to append
    def insert (self, next):
        if self.__Next != None:
            next.insert (self.__Next)
        self.__Next = next
    
    # -- remove the damage object immediately following this one
    def remove (self):
        if self.__Next != None:
            self.__Next = self.__Next.get_next ()

    # -- return the damage object immediately following this one     
    def get_next (self):
        return self.__Next


